from django.conf import settings
from django.contrib.auth.models import User
from recommends.providers import RecommendationProvider
from recommends.providers import recommendation_registry
try:
    from recommends.storages.mongodb.storage import MongoStorage
except ImportError:
    MongoStorage = None
try:
    from recommends.storages.redis.storage import RedisStorage
except ImportError:
    RedisStorage = None
try:
    from recommends.algorithms.pyrecsys import RecSysAlgorithm
except ImportError:
    RecSysAlgorithm = None
try:
    from recommends.tasks import recommends_precompute
except ImportError:
    recommends_precompute = None

if recommends_precompute is not None:
    from recommends.tests.tests import RecommendsTestCase

from recommends.tests.models import RecProduct, RecVote


class ProductRecommendationProvider(RecommendationProvider):

    def get_users(self):
        return User.objects.filter(is_active=True, rec_votes__isnull=False).distinct()

    def get_items(self):
        return RecProduct.objects.all()

    def get_ratings(self, obj):
        return RecVote.objects.filter(product=obj)

    def get_rating_score(self, rating):
        return rating.score

    def get_rating_site(self, rating):
        return rating.site

    def get_rating_user(self, rating):
        return rating.user

    def get_rating_item(self, rating):
        return rating.product

recommendation_registry.register(
    RecVote,
    [RecProduct],
    ProductRecommendationProvider)


class GhettoRecommendationProvider(RecommendationProvider):

    def get_users(self):
        return User.objects.filter(is_active=True, rec_votes__isnull=False).distinct()

    def get_items(self):
        return RecProduct.objects.all()

    def get_ratings(self, obj):
        return RecVote.objects.filter(product=obj)

    def get_rating_score(self, rating):
        return rating.score

    def get_rating_site(self, rating):
        return rating.site

    def get_rating_user(self, rating):
        return rating.user

    def get_rating_item(self, rating):
        return rating.product


if recommends_precompute is not None and RecSysAlgorithm is not None and getattr(settings, 'RECOMMENDS_TEST_RECSYS', False):
    class RecSysRecommendationProvider(ProductRecommendationProvider):
        algorithm = RecSysAlgorithm()

    class RecSysAlgoTestCase(RecommendsTestCase):
        results = {
            'len_recommended': 4,
            'len_similar_to_mug': 5
        }

        def setUp(self):
            recommendation_registry.unregister(
                RecVote,
                [RecProduct],
                ProductRecommendationProvider)
            recommendation_registry.register(
                RecVote,
                [RecProduct],
                RecSysRecommendationProvider)
            super(RecSysAlgoTestCase, self).setUp()

        def tearDown(self):
            super(RecSysAlgoTestCase, self).tearDown()
            recommendation_registry.unregister(
                RecVote,
                [RecProduct],
                RecSysRecommendationProvider)
            recommendation_registry.register(
                RecVote,
                [RecProduct],
                ProductRecommendationProvider)

if recommends_precompute is not None and RedisStorage is not None and getattr(settings, 'RECOMMENDS_TEST_REDIS', False):
    class RedisRecommendationProvider(GhettoRecommendationProvider):
        storage = RedisStorage(settings=settings)

    class RecommendsRedisStorageTestCase(RecommendsTestCase):

        def setUp(self):
            recommendation_registry.unregister(
                RecVote,
                [RecProduct],
                ProductRecommendationProvider)
            recommendation_registry.register(
                RecVote,
                [RecProduct],
                RedisRecommendationProvider)
            super(RecommendsRedisStorageTestCase, self).setUp()

        def tearDown(self):
            super(RecommendsRedisStorageTestCase, self).tearDown()
            recommendation_registry.unregister(
                RecVote,
                [RecProduct],
                RedisRecommendationProvider)
            recommendation_registry.register(
                RecVote,
                [RecProduct],
                ProductRecommendationProvider)

if recommends_precompute is not None and MongoStorage is not None and getattr(settings, 'RECOMMENDS_TEST_MONGO', False):
    class MongoRecommendationProvider(GhettoRecommendationProvider):
        storage = MongoStorage(settings=settings)

    class RecommendsMongoStorageTestCase(RecommendsTestCase):

        def setUp(self):
            recommendation_registry.unregister(
                RecVote,
                [RecProduct],
                ProductRecommendationProvider)
            recommendation_registry.register(
                RecVote,
                [RecProduct],
                MongoRecommendationProvider)
            super(RecommendsMongoStorageTestCase, self).setUp()

        def tearDown(self):
            super(RecommendsMongoStorageTestCase, self).tearDown()
            recommendation_registry.unregister(
                RecVote,
                [RecProduct],
                MongoRecommendationProvider)
            recommendation_registry.register(
                RecVote,
                [RecProduct],
                ProductRecommendationProvider)



try:
    from django.conf.urls.defaults import patterns, include, url
except ImportError:
    from django.conf.urls import patterns, include, url
from django.views.generic import TemplateView, DetailView
from .models import RecProduct


# Uncomment the next two lines to enable the admin:
from django.contrib import admin
admin.autodiscover()

urlpatterns = patterns('',
    # Examples:
    # url(r'^$', 'example_project.views.home', name='home'),
    # url(r'^example_project/',
    # include('example_project.foo.urls')),

    # Uncomment the admin/doc line below to enable admin documentation:
    # url(r'^admin/doc/',
    # include('django.contrib.admindocs.urls')),

    # Uncomment the next line to enable the admin:
    url(r'^admin/', include(admin.site.urls)),
    url(r'^login/',
        'recommends.tests.views.login', name='login'),
    url(r'^product/(?P<pk>\d+)/$',
        DetailView.as_view(
            model=RecProduct), name='product_detail'),
    url(r'^$', TemplateView.as_view(
        template_name='home.html'), name='home'),
)



from recsys.datamodel.data import Data
from recsys.algorithm.factorize import SVD
from .base import BaseAlgorithm
from recommends.converters import convert_vote_list_to_itemprefs


class RecSysAlgorithm(BaseAlgorithm):
    def __init__(self, k=100, *args, **kwargs):
        self.k = k
        super(RecSysAlgorithm, self).__init__(*args, **kwargs)

    @property
    def svd(self):
        return self.cache.get('svd', None)

    def setup_svd(self, vote_list):
        if self.svd is None:
            self.cache['svd'] = SVD()
            data = Data()

            for vote in vote_list:
                user_id = vote[0].id
                item_id = vote[1]
                value = float(vote[2])
                data.add_tuple((value, item_id, user_id))  # Tuple format is: <value, row, column>
            self.cache['svd'].set_data(data)
            self.cache['svd'].compute(k=self.k, min_values=1)
        return self.svd

    def calculate_similarities(self, vote_list, verbose=0):
        svd = self.setup_svd(vote_list)

        itemPrefs = convert_vote_list_to_itemprefs(vote_list)
        itemMatch = {}
        for item in itemPrefs:
            itemMatch[item] = svd.similar(item)
        iteritems = itemMatch.items()
        return iteritems

    def calculate_recommendations(self, vote_list, itemMatch, itemIgnored):
        svd = self.setup_svd(vote_list)

        recommendations = []
        users = set(map(lambda x: x[0], vote_list))
        for user in users:
            try:
                rankings = svd.recommend(user.id, only_unknowns=True, is_row=False)
                recommendations.append((user, rankings))
            except KeyError:
                pass
        return recommendations



import logging
import math
import redis
from recommends.models import MockModel, MockSimilarity
from recommends.storages.base import BaseRecommendationStorage
from recommends.settings import RECOMMENDS_LOGGER_NAME, RECOMMENDS_STORAGE_LOGGING_THRESHOLD
from .settings import RECOMMENDS_STORAGE_REDIS_DATABASE
from .managers import RedisStorageManager


logger = logging.getLogger(RECOMMENDS_LOGGER_NAME)


class RedisStorage(BaseRecommendationStorage):
    manager = RedisStorageManager()

    _redis = None

    can_lock = True
    LOCK_KEY = 'recommends-redis-lock-key'

    @property
    def redis(self):
        if self._redis is None:
            self._redis = redis.StrictRedis(host=RECOMMENDS_STORAGE_REDIS_DATABASE['HOST'], port=RECOMMENDS_STORAGE_REDIS_DATABASE['PORT'], db=RECOMMENDS_STORAGE_REDIS_DATABASE['NAME'])
        return self._redis

    def get_lock(self):
        return self.redis.setnx(self.LOCK_KEY, 1)

    def release_lock(self):
        return self.redis.delete(self.LOCK_KEY)

    def _get_mock_models(self, dicts, mock_class=MockModel):
        return map(lambda x: mock_class(**x), dicts)

    def get_similarities_for_object(self, obj, limit=10, raw_id=False):
        r = self.redis

        object_id = self.get_identifier(obj)
        key = 'recommends:similarity:%s' % object_id
        scores = r.zrevrangebyscore(key, min=0, max=1, num=limit, start=0, withscores=True)

        similarity_dicts = []
        for identifier, score in scores:
            similarity_dict = self.identifier_manager.identifier_to_dict(object_id)
            similarity_dict.update(self.identifier_manager.identifier_to_dict(identifier, score, related=True))
            similarity_dicts.append(similarity_dict)
        if raw_id:
            return [{
                'related_object_id': item['related_object_id'],
                'contect_type_id': item['object_ctype']}
                for item in similarity_dicts][:limit]
        return self._get_mock_models(similarity_dicts, mock_class=MockSimilarity)

    def get_recommendations_for_user(self, user, limit=10, raw_id=False):
        r = self.redis
        key = 'recommends:recommendation:%s' % user.id
        scores = r.zrevrangebyscore(key, min=0, max=1, num=limit, start=0, withscores=True)

        recommendation_dicts = [self.identifier_manager.identifier_to_dict(object_id, score) for object_id, score in scores]
        if raw_id:
            return [{
                    'object_id': item['object_id'],
                    'contect_type_id': item['object_ctype']}
                    for item in recommendation_dicts][:limit]
        return self._get_mock_models(recommendation_dicts, mock_class=MockModel)

    def get_votes(self):
        pass

    def store_votes(self, iterable):
        pass

    def store_similarities(self, itemMatch):
        r = self.redis

        logger.info('saving similarities')
        count = 0

        for object_id, scores in itemMatch:
            object_target, object_target_site = self.resolve_identifier(object_id)

            for related_object_id, score in scores:
                if not math.isnan(score) and score > self.threshold_similarities:
                    object_related, object_related_site = self.resolve_identifier(related_object_id)
                    if object_target != object_related:
                        key = 'recommends:similarity:%s' % object_id
                        r.zadd(key, score, related_object_id)

                        rev_key = 'recommends:similarity_reverse:%s' % related_object_id
                        r.sadd(rev_key, object_id)

                        index_key = 'recommends:similarity:index'
                        r.sadd(index_key, object_id)

                        count = count + 1

                        if count % RECOMMENDS_STORAGE_LOGGING_THRESHOLD == 0:
                            logger.debug('saved %s similarities...' % count)
        logger.info('saved %s similarities...' % count)

    def store_recommendations(self, recommendations):
        r = self.redis

        logger.info('saving recommendation')
        count = 0

        for (user, rankings) in recommendations:
            for object_id, score in rankings:
                if not math.isnan(score) and score > self.threshold_recommendations:
                    key = 'recommends:recommendation:%s' % user.id
                    r.zadd(key, score, object_id)

                    rev_key = 'recommends:recommendation_reverse:%s' % object_id
                    r.zadd(rev_key, score, user.id)

                    index_key = 'recommends:recommendation:index'
                    r.sadd(index_key, object_id)

                    count = count + 1
                    if count % RECOMMENDS_STORAGE_LOGGING_THRESHOLD == 0:
                        logger.debug('saved %s recommendations...' % count)
        logger.info('saved %s recommendation...' % count)

    def remove_similarities(self, obj):
        r = self.redis
        object_id = self.get_identifier(obj)
        key = 'recommends:similarity:%s' % object_id
        r.delete(key)

        rev_key = 'recommends:similarity_reverse:%s' % object_id
        values = r.smembers(rev_key)
        for value in values:
            r.zrem('recommends:similarity:%s' % value, object_id)
        r.delete(rev_key)

        index_key = 'recommends:similarity:index'
        r.srem(index_key, object_id)

    def remove_recommendations(self, obj):
        r = self.redis
        object_id = self.get_identifier(obj)

        rev_key = 'recommends:recommendation_reverse:%s' % object_id
        user_ids = r.zrevrangebyscore(rev_key, min=0, max=1, start=0, num=r.zcount(rev_key, min=0, max=1))
        for user_id in user_ids:
            r.zrem('recommends:recommendation:%s' % user_id, object_id)
        r.delete(rev_key)

        index_key = 'recommends:recommendation:index'
        r.srem(index_key, object_id)



class BaseAlgorithm(object):
    """
    """

    _cache = {}

    def clear_cache(self):
        self._cache = {}

    @property
    def cache(self):
        return self._cache

    def calculate_similarities(self, vote_list, verbose=0):
        """
        Must return an dict of similarities for every object:

        Accepts a vote matrix representing votes with the following schema:

        ::

            [
                (<user1>, "<object_identifier1>", <score>),
                (<user1>, "<object_identifier2>", <score>),
            ]

        Output must be a dictionary with the following schema:

        ::

            [
                ("<object_identifier1>", [
                                ("<related_object_identifier2>", <score>),
                                ("<related_object_identifier3>", <score>),
                ]),
                ("<object_identifier2>", [
                                ("<related_object_identifier1>", <score>),
                                ("<related_object_identifier3>", <score>),
                ]),
            ]

        """
        raise NotImplemented

    def calculate_recommendations(self, vote_list, itemMatch):
        """
        ``itemMatch`` is supposed to be the result of ``calculate_similarities()``

        Returns a list of recommendations:

        ::

            [
                (<user1>, [
                    ("<object_identifier1>", <score>),
                    ("<object_identifier2>", <score>),
                ]),
                (<user2>, [
                    ("<object_identifier1>", <score>),
                    ("<object_identifier2>", <score>),
                ]),
            ]

        """
        raise NotImplemented



import timeit
from django.test import TestCase
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.test import Client
from recommends.providers import recommendation_registry
from recommends.tasks import recommends_precompute
from .models import RecProduct, RecVote
from django.test.utils import override_settings
import os


@override_settings(CELERY_DB_REUSE_MAX=200, LANGUAGES=(
    ('en', 'English'),),
    LANGUAGE_CODE='en',
    TEMPLATE_DIRS=(
        os.path.join(os.path.dirname(__file__), 'templates'),),
    TEMPLATE_LOADERS=('django.template.loaders.filesystem.Loader',),
    USE_TZ=False, )
class RecommendsTestCase(TestCase):
    fixtures = ['products.json']
    urls = 'recommends.tests.urls'

    results = {
        'len_recommended': 2,
        'len_similar_to_mug': 2
    }

    def setUp(self):
        self.client = Client()
        self.mug = RecProduct.objects.get(name='Coffee Mug')
        self.orange_juice = RecProduct.objects.get(name='Orange Juice')
        self.wine = RecProduct.objects.get(name='Bottle of Red Wine')
        RecProduct.objects.get(name='1lb Tenderloin Steak').delete()
        self.user1 = User.objects.get(username='user1')

        from django.template import loader
        loader.template_source_loaders = None

        self.provider = recommendation_registry.get_provider_for_content(RecProduct)
        recommends_precompute()

    def tearDown(self):
        from django.template import loader
        loader.template_source_loaders = None
        super(RecommendsTestCase, self).tearDown()

    def isObjectWithIdExists(self, object_id):
        return RecProduct.objects.filter(id=object_id).exists()

    def test_similarities(self):
        # test similarities objects
        similar_to_mug = self.provider.storage.get_similarities_for_object(self.mug)
        self.assertNotEquals(len(similar_to_mug), 0)
        self.assertEquals(len(similar_to_mug), self.results['len_similar_to_mug'])
        self.assertTrue(self.wine in [s.related_object for s in similar_to_mug])
        # Make sure we didn't get all 0s
        zero_scores = filter(lambda x: x.score == 0, similar_to_mug)
        self.assertNotEquals(len(zero_scores), len(similar_to_mug))

    def test_similarities_raw_ids(self):
        # test similarities raw ids
        similar_to_mug_ids = self.provider.storage.get_similarities_for_object(self.mug, raw_id=True)
        self.assertNotEquals(len(similar_to_mug_ids), 0)
        self.assertEquals(len(similar_to_mug_ids), self.results['len_similar_to_mug'])
        similar_to_mug_related_ids = [item['related_object_id'] for item in similar_to_mug_ids]
        self.assertTrue(self.wine.id in similar_to_mug_related_ids)
        self.assertTrue(all([self.isObjectWithIdExists(related_object_id) for related_object_id in similar_to_mug_related_ids]))

    def test_recommendation(self):
        # test recommendations
        recommendations = self.provider.storage.get_recommendations_for_user(self.user1)
        self.assertNotEquals(len(recommendations), 0)
        self.assertEquals(len(recommendations), self.results['len_recommended'])
        self.assertTrue(self.wine in [s.object for s in recommendations])
        # Make sure we didn't get all 0s
        zero_scores = filter(lambda x: x.score == 0, recommendations)
        self.assertNotEquals(len(zero_scores), len(recommendations))
        # Make sure we don't recommend item that the user already have
        self.assertFalse(self.mug in [v.product for v in RecVote.objects.filter(user=self.user1)])

    def test_recommendation_raw_ids(self):
        # test recommendation raw ids
        recommendation_ids = self.provider.storage.get_recommendations_for_user(self.user1, raw_id=True)
        self.assertNotEquals(len(recommendation_ids), 0)
        self.assertEquals(len(recommendation_ids), self.results['len_recommended'])
        recommendation_object_ids = [item['object_id'] for item in recommendation_ids]
        self.assertTrue(self.wine.id in recommendation_object_ids)
        self.assertTrue(all([self.isObjectWithIdExists(object_id) for object_id in recommendation_object_ids]))

    def test_views(self):
        self.client.login(username='user1', password='user1')

        response = self.client.get(self.mug.get_absolute_url())
        self.assertTrue(self.orange_juice.get_absolute_url() in response.content)

    def _test_performance(self):
        stmt = """recommends_precompute()"""
        setup = """from recommends.tasks import recommends_precompute"""
        print "timing..."
        times = timeit.repeat(stmt, setup, number=100)
        print times


class RecommendsListenersTestCase(TestCase):
    fixtures = ['products.json']

    def setUp(self):
        self.client = Client()
        self.mug = RecProduct.objects.get(name='Coffee Mug')
        self.orange_juice = RecProduct.objects.get(name='Orange Juice')
        self.wine = RecProduct.objects.get(name='Bottle of Red Wine')
        self.steak = RecProduct.objects.get(name='1lb Tenderloin Steak')
        self.user1 = User.objects.get(username='user1')

        self.provider = recommendation_registry.get_provider_for_content(RecProduct)

        recommends_precompute()

    def test_listeners(self):
        self.client.login(username='user1', password='user1')
        response = self.client.get(reverse('home'))

        self.vote = RecVote.objects.create(
            user=self.user1,
            product=self.steak,
            site_id=1,
            score=1
        )

        response = self.client.get(reverse('home'))
        steak_url = self.steak.get_absolute_url()
        self.assertTrue(steak_url in response.content)
        recommendations = self.provider.storage.get_recommendations_for_user(self.user1)
        steak_recs = filter(lambda x: x.object_id == self.steak.id, recommendations)
        self.assertEqual(1L, len(steak_recs))

        self.steak.delete()

        response = self.client.get(reverse('home'))
        self.assertFalse(steak_url in response.content)
        recommendations = self.provider.storage.get_recommendations_for_user(self.user1)
        steak_recs = filter(lambda x: x.object_id == self.steak.id, recommendations)
        self.assertEqual(0L, len(steak_recs))

    def tearDown(self):
        self.vote.delete()
        recommends_precompute()



from django.core.management.base import BaseCommand
from recommends.tasks import recommends_precompute

from datetime import datetime
import dateutil.relativedelta
from optparse import make_option

import warnings


class Command(BaseCommand):
    help = 'Calculate recommendations and similarities based on ratings'
    option_list = BaseCommand.option_list + (
        make_option('--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='verbose mode'
        ),
    )

    def handle(self, *args, **options):
        verbosity = int(options.get('verbosity', 0))
        if options['verbose']:
            warnings.warn('The `--verbose` option is being deprecated and it will be removed in the next release. Use `--verbosity` instead.', PendingDeprecationWarning)
            verbosity = 1

        if verbosity == 0:
            # avoids allocating the results
            recommends_precompute()
        else:
            self.stdout.write("\nCalculation Started.\n")
            start_time = datetime.now()
            results = recommends_precompute()
            end_time = datetime.now()
            if verbosity > 1:
                for r in results:
                    self.stdout.write(
                        "%d similarities and %d recommendations saved.\n"
                        % (r['similar_count'], r['recommend_count']))
            rd = dateutil.relativedelta.relativedelta(end_time, start_time)
            self.stdout.write(
                "Calculation finished in %d years, %d months, %d days, %d hours, %d minutes and %d seconds\n"
                % (rd.years, rd.months, rd.days, rd.hours, rd.minutes, rd.seconds))



from celery.task import task, periodic_task
from celery.schedules import crontab
from .utils import filelock

from .settings import RECOMMENDS_TASK_RUN, RECOMMENDS_TASK_CRONTAB, RECOMMENDS_TASK_EXPIRES


def recommends_precompute():
    results = []
    from .providers import recommendation_registry

    # I know this is weird, but it's faster (tested on CPyhton 2.6.5)
    def _precompute(provider_instance):
        results.append(provider_instance.precompute())

    if recommendation_registry.storage.can_lock:
        locked = recommendation_registry.storage.get_lock()
        if locked:
            [_precompute(provider_instance) for provider_instance in recommendation_registry.get_vote_providers()]
            recommendation_registry.storage.release_lock()
    else:
        with filelock('recommends_precompute.lock'):
            [_precompute(provider_instance)
             for provider_instance in recommendation_registry.get_vote_providers()]

    return results

if RECOMMENDS_TASK_RUN:
    @periodic_task(name='recommends_precompute', run_every=crontab(**RECOMMENDS_TASK_CRONTAB), expires=RECOMMENDS_TASK_EXPIRES)
    def _recommends_precompute():
        recommends_precompute()


@task(name='remove_suggestions')
def remove_suggestions(rated_model, object_id):
    from django.db.models import get_model
    from recommends.providers import recommendation_registry

    ObjectClass = get_model(*rated_model.split('.'))
    provider_instance = recommendation_registry.get_provider_for_content(
        ObjectClass)
    obj = ObjectClass.objects.get(pk=object_id)

    provider_instance.storage.remove_recommendations(obj)


@task(name='remove_similarities')
def remove_similarities(rated_model, object_id):
    from django.db.models import get_model
    from recommends.providers import recommendation_registry

    ObjectClass = get_model(*rated_model.split('.'))
    provider_instance = recommendation_registry.get_provider_for_content(
        ObjectClass)
    obj = ObjectClass.objects.get(pk=object_id)

    provider_instance.storage.remove_similarities(obj)



from .settings import RECOMMENDS_STORAGE_DATABASE_ALIAS


class RecommendsRouter(object):
    def db_for_read(self, model, **hints):
        "Point all operations on recommends models to RECOMMENDS_STORAGE_DATABASE_NAME"
        if model._meta.app_label == 'recommends':
            return RECOMMENDS_STORAGE_DATABASE_ALIAS
        return None

    def db_for_write(self, model, **hints):
        "Point all operations on recommends models to RECOMMENDS_STORAGE_DATABASE_NAME"
        if model._meta.app_label == 'recommends':
            return RECOMMENDS_STORAGE_DATABASE_ALIAS
        return None

    def allow_relation(self, obj1, obj2, **hints):
        "Allow any relation if a model in recommends is involved"
        if obj1._meta.app_label == 'recommends' or obj2._meta.app_label == 'recommends':
            return True
        return None

    def allow_syncdb(self, db, model):
        "Make sure the recommends app only appears on the RECOMMENDS_STORAGE_DATABASE_NAME db"
        if db == RECOMMENDS_STORAGE_DATABASE_ALIAS:
            return model._meta.app_label == 'recommends'
        elif model._meta.app_label == 'recommends':
            return False
        return None



from ..providers import recommendation_registry
from ..settings import RECOMMENDS_CACHE_TEMPLATETAGS_TIMEOUT
from django import template
from django.core.cache import cache
from django.conf import settings
from django.db import models
register = template.Library()


@register.filter
def similarities(obj, limit=5):
    """
    Returns a list of Similarity objects, representing how much an object is similar to the given one.

    Usage:

    ::

        {% for similarity in myobj|similar:5 %}
            {{ similarity.related_object }}
        {% endfor %}
    """
    if isinstance(obj, models.Model):
        cache_key = 'recommends:similarities:%s:%s.%s:%s:%s' % (settings.SITE_ID, obj._meta.app_label, obj._meta.object_name.lower(), obj.id, limit)
        similarities = cache.get(cache_key)
        if similarities is None:
            provider = recommendation_registry.get_provider_for_content(obj)
            similarities = provider.storage.get_similarities_for_object(obj, int(limit))
            cache.set(cache_key, similarities, RECOMMENDS_CACHE_TEMPLATETAGS_TIMEOUT)
        return similarities


class SuggestionNode(template.Node):
    def __init__(self, varname, limit):
        self.varname = varname
        self.limit = limit

    def render(self, context):
        user = context['user']
        if user.is_authenticated():  # We need an id after all
            cache_key = 'recommends:recommendations:%s:%s:%s' % (settings.SITE_ID, user.id, self.limit)
            suggestions = cache.get(cache_key)
            if suggestions is None:
                suggestions = set()
                for provider in recommendation_registry.providers:
                    suggestions.update(provider.storage.get_recommendations_for_user(user, int(self.limit)))
                cache.set(cache_key, suggestions, RECOMMENDS_CACHE_TEMPLATETAGS_TIMEOUT)
            context[self.varname] = suggestions
        return ''


@register.tag()
def suggested(parser, token):
    """
    Returns a list of Recommendation (suggestions of objects) for the current user.

    Usage:

    ::

        {% suggested as suggestions [limit 5]  %}
        {% for suggested in suggestions %}
            {{ suggested.object }}
        {% endfor %}
    """
    bits = token.contents.split()
    varname = bits[2]
    try:
        limit = int(bits[4])
    except IndexError:
        limit = 5
    return SuggestionNode(varname, limit)



import os
from setuptools import setup, find_packages

VERSION = "0.0.20"


def read(fname):
    try:
        with open(os.path.join(os.path.dirname(__file__), fname)) as fh:
            return fh.read()
    except IOError:
        return ''

requirements = read('requirements').splitlines()
tests_requirements = read('test-requirements').splitlines()

setup(
    name="django-recommends",
    version=VERSION,
    description="A django app that builds item-based suggestions for users",
    long_description=read('README.rst'),
    url='https://github.com/python-recsys/django-recommends',
    license='MIT',
    author='Flavio Curella',
    author_email='flavio.curella@gmail.com',
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
    install_requires=requirements,
    tests_require=tests_requirements,
)



from django.conf import settings
from django.db import models
from recommends.managers import CachedContentTypesMixin


class RecommendsManager(models.Manager, CachedContentTypesMixin):
    def filter_for_model(self, model):
        ctype_id = self.get_ctype_id_for_obj(model)
        return self.filter(object_ctype=ctype_id)

    def filter_for_object(self, obj):
        return self.filter_for_model(obj).filter(object_id=obj.id)


class SimilarityManager(RecommendsManager):
    def filter_for_related_model(self, related_model):
        ctype_id = self.get_ctype_id_for_obj(related_model)
        return self.filter(related_object_ctype=ctype_id)

    def filter_for_related_object(self, related_obj):
        return self.filter_for_related_model(related_obj).filter(related_object_id=related_obj.id)

    def filter_by_couple(self, target_object, related_obj):
        related_ctype_id = self.get_ctype_id_for_obj(related_obj)

        return self.filter_for_object(target_object).filter(
            related_object_ctype=related_ctype_id,
            related_object_id=related_obj.id
        )

    def get_query_set(self):
        return super(SimilarityManager, self).get_query_set().filter(score__isnull=False)

    def get_or_create_for_objects(self, object_target, object_target_site, object_related, object_related_site):
        object_ctype_id = self.get_ctype_id_for_obj(object_target)
        object_id = object_target.id

        related_object_ctype_id = self.get_ctype_id_for_obj(object_related)
        related_object_id = object_related.id

        return self.get_or_create(
            object_ctype=object_ctype_id,
            object_id=object_id,
            object_site=object_target_site.id,
            related_object_ctype=related_object_ctype_id,
            related_object_id=related_object_id,
            related_object_site=object_related_site.id
        )

    def set_score_for_objects(self, object_target, object_target_site, object_related, object_related_site, score):
        if score == 0:
            self.filter_by_couple(object_target, object_related).filter(
                object_site=object_target_site.id,
                related_object_site=object_related_site.id
            ).delete()
            return None

        result, created = self.get_or_create_for_objects(object_target, object_target_site, object_related, object_related_site)
        result.score = score
        result.save()
        return result

    def similar_to(self, obj, site=None, **kwargs):
        if site is None and 'related_object_site' not in kwargs:
            kwargs['related_object_site'] = settings.SITE_ID
        return self.filter_for_object(obj).filter(**kwargs)


class RecommendationManager(RecommendsManager):
    def get_query_set(self):
        return super(RecommendationManager, self).get_query_set().filter(score__isnull=False)

    def get_or_create_for_object(self, user, object_recommended, object_site):
        object_ctype_id = self.get_ctype_id_for_obj(object_recommended)
        object_id = object_recommended.id

        return self.get_or_create(
            object_ctype=object_ctype_id,
            object_id=object_id,
            object_site=object_site.id,
            user=user.id
        )

    def set_score_for_object(self, user, object_recommended, object_site, score):
        if score == 0:
            self.filter_for_object(object_recommended).filter(user=user.id).delete()
            return None

        result, created = self.get_or_create_for_object(user, object_recommended, object_site)
        result.score = score
        result.save()
        return result
